# Imports.
# Local imports.

# External imports.
import ezdxf
from ezdxf.gfxattribs import GfxAttribs
from ezdxf.document import Drawing
from ezdxf.transform import inplace, scale
from ezdxf.math import Matrix44
from ezdxf.enums import TextEntityAlignment
from ezdxf.tools.standards import linetypes

from math import pi

# Classes references.
doc_class = Drawing
attrib_class = GfxAttribs

linetypes_list = linetypes(scale=0.1)
lt_center = linetypes_list[1]

# Matrix transformation definition.
matrix_x_mirror = Matrix44([1, 0, 0, 0],
                           [0, -1, 0, 0],
                           [0, 0, 1, 0],
                           [0, 0, 0, 1])

matrix_y_mirror = Matrix44([-1, 0, 0, 0],
                           [0, 1, 0, 0],
                           [0, 0, 1, 0],
                           [0, 0, 0, 1])


# Function that draws a circunference.
def circle(doc: Drawing, center_point: tuple, radius: float, attr=None):

    if not isinstance(doc, Drawing):
        return []

    msp = doc.modelspace()

    # Drawing of circle.
    circ = [msp.add_circle(center_point, radius=radius, dxfattribs=attr)]

    return circ


# Function that draws a curve with/out thickness.
def curve(doc: Drawing, center_point: tuple, radius: float, start_angle: float, end_angle: float, thickness: float = 0,
          attr=None) -> list:

    if not isinstance(doc, Drawing):
        return []

    msp = doc.modelspace()

    # Drawing of arc.
    arcs = [msp.add_arc(center_point, radius=radius, start_angle=start_angle, end_angle=end_angle, dxfattribs=attr)]

    if thickness:
        arcs += [msp.add_arc(center_point, radius=radius + thickness, start_angle=start_angle, end_angle=end_angle,
                             dxfattribs=attr)]

    return arcs


# Create a delimit axe and rerturns and list with elements.
def delimit_axe(document: Drawing, x: float, y: float, height: float = 1, radius: float = 0.1, text_height: float = 0.1,
                symbol: str = None, direction: int = 1, attr: GfxAttribs = None) -> list:

    document.linetypes.add(name=lt_center[0], description=lt_center[1], pattern=lt_center[2])
    msp = document.modelspace()

    if direction:
        group = [msp.add_line(start=(x, y), end=(x, y + height), dxfattribs=attr)]
        if symbol is not None:
            group += [msp.add_circle(center=(x, y + height + radius), radius=radius)]
            group += [msp.add_text(text=symbol, height=text_height).set_placement(
                p1=(x, y + height + radius), align=TextEntityAlignment.MIDDLE_CENTER)]
    else:
        group = [msp.add_line(start=(x + radius, y), end=(x + radius + height, y), dxfattribs=attr)]
        if symbol is not None:
            group += [msp.add_circle(center=(x, y), radius=radius)]
            group += [msp.add_text(text=symbol, height=text_height).set_placement(
                p1=(x, y), align=TextEntityAlignment.MIDDLE_CENTER)]

    return group


# Draws a linear dimension.
def dim_linear(document: Drawing, p_base: tuple, p1: tuple, p2: tuple, rotation: float = 0,
               dimstyle: str = "EZ_M_25_H25_CM", attr: GfxAttribs = None) -> list:
    msp = document.modelspace()

    dims = [msp.add_linear_dim(base=p_base, p1=p1, p2=p2, dimstyle=dimstyle, angle=rotation, dxfattribs=attr)]
    dims[-1].render()

    return dims


# Returns the a and b constants of the line equation from two poins.
def get_line_eq(x1, y1, x2, y2):
    if x1 == x2:  # For vertical lines.
        a = float(9999999999)  # Aproximate infinite.
        b = - a * x1

    else:  # For not vertical lines.
        a = (y2 - y1) / (x2 - x1)
        b = y1 - a * x1

    return a, b


# Returns the intersection of two lines.
def get_lines_intersec(a1: float, b1: float, a2: float, b2: float) -> float:

    x = (b2 - b1) / (a1 - a2)

    return x


# Function that draws a line.
def line(doc: Drawing, p1: tuple, p2: tuple, attr=None) -> list:

    line_segment = doc.modelspace().add_line(p1, p2, dxfattribs=attr)

    return [line_segment]


# Function that mirrors a given group of objects.
def mirror(objects: list, mirror_type: str, in_situ: bool = False) -> int:
    # Axis "x" mirror.
    if "x" in mirror_type:
        inplace(objects, matrix_x_mirror)

    # Axis "y" mirror.
    if "y" in mirror_type:
        inplace(objects, matrix_y_mirror)

    return 1


# Returns the points ordered in function of X coordinate.
def point_order(x1: float, y1: float, x2: float, y2: float) -> tuple[float, float, float, float]:
    if x1 <= x2:
        return x1, y1, x2, y2
    else:
        return x2, y2, x1, y1


# Function that transform degrees to radians.
def rads(degrees: float) -> float:
    return degrees * pi / 180


# Function that draws a rectangle in dxf file given.
def rect(doc: Drawing, width: float, height: float, x: float, y: float, sides=None, fill: bool = False,
         polyline: bool = False, attr=None) -> list:

    if sides is None:
        sides = [1, 1, 1, 1]

    if not isinstance(doc, Drawing):
        return []

    msp = doc.modelspace()
    lines = []

    # Polyline rectangle.
    if polyline:
        lines.append(msp.add_lwpolyline([(x, y), (x + width, y), (x + width, y + height), (x, y + height), (x, y)],
                                        dxfattribs=attr))
    else:
        # Horizontal lines.
        if sides[0]:
            lines.append(msp.add_line((x, y + height), (x + width, y + height), dxfattribs=attr))  # Top side.
        if sides[2]:
            lines.append(msp.add_line((x, y), (x + width, y), dxfattribs=attr))  # Bottom side.

        # Vertical lines.
        if sides[1]:
            lines.append(msp.add_line((x + width, y), (x + width, y + height), dxfattribs=attr))  # Right side.
        if sides[3]:
            lines.append(msp.add_line((x, y), (x, y + height), dxfattribs=attr))  # Left side.

    if fill:
        hatch = msp.add_hatch(color=7, dxfattribs=attr)
        hatch.paths.add_polyline_path([(x, y), (x + width, y), (x + width, y + height), (x, y + height)])
        hatch.set_pattern_fill("ANSI31", scale=0.01)
        lines.append(hatch)

    return lines


# Function that draws a rectangle in dxf file given.
def rect_border_curve(doc: Drawing, width: float, height: float, radius: float, x: float, y: float,
                      thickness: float = 0, sides=None, curves=None, curves_radius=None, attr=None) -> list:

    if sides is None:
        sides = [1, 1, 1, 1]

    if curves is None:
        curves = [1, 1, 1, 1]

    if curves_radius is None:
        curves_radius = [radius, radius, radius, radius]

    if not isinstance(doc, Drawing):
        return []

    msp = doc.modelspace()
    lines = []

    # Horizontal lines.
    if sides[0]:
        lines.append(msp.add_line(start=(x + curves_radius[0], y + height),
                                  end=(x - curves_radius[1] + width, y + height),  dxfattribs=attr))  # Top side.
    if sides[2]:
        lines.append(msp.add_line(start=(x + curves_radius[3], y),
                                  end=(x - curves_radius[2] + width, y), dxfattribs=attr))  # Bottom side.

    # Vertical lines.
    if sides[1]:
        lines.append(msp.add_line(start=(x + width, y + curves_radius[2]),
                                  end=(x + width, y - curves_radius[1] + height), dxfattribs=attr))  # Right side.
    if sides[3]:
        lines.append(msp.add_line(start=(x, y + curves_radius[3]),
                                  end=(x, y - curves_radius[0] + height), dxfattribs=attr))  # Left side.

    if thickness:
        # Horizontal lines borders.
        if sides[0]:
            lines.append(msp.add_line(start=(x + curves_radius[0], y + thickness + height),
                                      end=(x - curves_radius[1] + width, y + thickness + height),
                                      dxfattribs=attr))  # Top side.
        if sides[2]:
            lines.append(msp.add_line(start=(x + curves_radius[3], y - thickness),
                                      end=(x - curves_radius[2] + width, y - thickness),
                                      dxfattribs=attr))  # Bottom side.

        # Vertical lines borders.
        if sides[1]:
            lines.append(msp.add_line(start=(x + thickness + width, y + curves_radius[2]),
                                      end=(x + thickness + width, y - curves_radius[1] + height),
                                      dxfattribs=attr))  # Right side.
        if sides[3]:
            lines.append(msp.add_line(start=(x - thickness, y + curves_radius[3]),
                                      end=(x - thickness, y - curves_radius[0] + height),
                                      dxfattribs=attr))  # Left side.

    # Curves.
    if (sides[0] or sides[3]) and curves[0]:
        lines += curve(doc=doc, center_point=(x + curves_radius[0], y - curves_radius[0] + height),
                       start_angle=90, end_angle=180, radius=curves_radius[0], thickness=thickness, attr=attr)
    if (sides[0] or sides[1]) and curves[1]:
        lines += curve(doc=doc, center_point=(x - curves_radius[1] + width, y - curves_radius[1] + height), start_angle=0, end_angle=90,
                       radius=curves_radius[1], thickness=thickness, attr=attr)
    if (sides[1] or sides[2]) and curves[2]:
        lines += curve(doc=doc, center_point=(x - curves_radius[2] + width, y + curves_radius[2]), start_angle=270, end_angle=360,
                       radius=curves_radius[2], thickness=thickness, attr=attr)
    if (sides[2] or sides[3]) and curves[3]:
        lines += curve(doc=doc, center_point=(x + curves_radius[3], y + curves_radius[3]), start_angle=180, end_angle=270,
                       radius=curves_radius[3], thickness=thickness, attr=attr)

    return lines


# Function that rotates elements.
def rotate(objects: list, angle: float) -> int:

    ezdxf.transform.z_rotate(entities=objects, angle=angle)

    return 1


# Function that draws simple text.
def text(document: Drawing, text: str, height: float, point: tuple, rotation: float = 0, attr = None) -> list:
    msp = document.modelspace()

    group = msp.add_text(text=text, height=height, rotation=rotation, dxfattribs=attr).set_placement(point)

    return [group]


# Function that translates a group of elements.
def translate(objects: list, vector: tuple) -> int:

    ezdxf.transform.translate(entities=objects, offset=vector)

    return 1



class Element:
    """
    General proposes for Element objects.

    :param width: Width of the Element.
    :type width: float
    :param height: Height of the Element.
    :type height: float
    :param type: Type of the Element (bar, stirrup, beam, etc.).
    :type type: str
    :param x: X coordinate of the bottom left corner bounding box Element point.
    :type x: float
    :param y: Y coordinate of the bottom left corner bounding box Element point.

    :ivar width: Width of the Element.
    :ivar height: Height of the Element.
    :ivar type: Type of the Element (bar, stirrup, beam, etc.).
    :ivar x: X coordinate of the bottom left corner bounding box Element point.
    :ivar y: Y coordinate of the bottom left corner bounding box Element point.
    """

    def __init__(self, width: float, height: float, type: str, x: float, y: float):
        """
        Initialeze a new instance of Element.
        """
        self.width = width
        self.height = height
        self.type = type
        self.x = x
        self.y = y



# Imports.
# Local imports.
from drawing_utils import (circle, curve, dim_linear, line, mirror, rads,
                           rect, rotate, text, translate)
from element import Element
from globals import STEEL_WEIGHT

# External imports.
import ezdxf
from ezdxf.document import Drawing
from math import cos, sin, tan, pi

# Constants.
SIN45, COS45 = sin(rads(45)), cos(rads(45))


class Bar(Element):
    """
    Longitudinal bar element, computes geometrics and physics props and manages dxf drawing methods.

        :param reinforcement_length: Length of the stirrup reinforcement.
        :type reinforcement_length: float
        :param diameter: Diameter of stirrup bar.
        :type diameter: float
        :param x: X coordinate of the bottom left corner Stirrup bounding box point.
        :type x: float
        :param y: Y coordinate of the bottom left corner Stirrup bounding box point.
        :type y: float
        :param left_anchor: Left anchor length of stirrup.
        :type left_anchor: float
        :param right_anchor: Right anchor length of stirrup.
        :type right_anchor: float
        :param mandrel_radius: Mandrel radius of bar.
        :type mandrel_radius: float
        :param direction: Direction of the stirrup (Horizontal or Vertical).
        :type direction: str
        :param orientation: Orientation of the stirrup (top, right, down, left).
        :type orientation: str
        :param bend_longitud: Bending longitud of the bar at center of bar.
        :type: float
        :param bend_angle: Bending declination angle.
        :type bend_angle: float
        :param bend_height: Bending height of the bar at center of bar.
        :type bend_height: float
        :param transverse_center: Transverse center of the drawing of cross-section.
        :type transverse_center: tuple

        :ivar bend_longitud: Bending longitud of the bar at center of bar.
        :ivar bend_angle: Bending declination angle.
        :ivar bend_height: Bending height of the bar at center of bar.
        :ivar bending_proyection: Proyection of the bend bar.
        :ivar left_anchor: Left anchor length of stirrup.
        :ivar right_anchor: Right anchor length of stirrup.
        :ivar mandrel_radius: Mandrel radius of bar.
        :ivar mandrel_radius_ext: Mandrel radius plus diameter of bar.
        :ivar diameter: Diameter of bar.
        :ivar radius: Radius of bar.
        :ivar direction: Direction of the bar (horizontal or vertical).
        :ivar orientation: Orientation of the bar (top, right, down, left).
        :ivar transverse_center: Transverse center of the drawing of cross-section.
        :ivar reinforcement_length: Length of the reinforcement.
        ivar length: Length of the bar.
        :ivar weight: Weight of the reinforcement, considering 7850 kg / m3.
        :ivar box_width: Width of the box that contains the bar.
        :ivar box_height: Height of the box that contains the bar.
    """
    def __init__(self, reinforcement_length: float, diameter: float, x: float = 0, y: float = 0, left_anchor: float = 0,
                 right_anchor: float = 0, mandrel_radius: float = 0, direction: str = "horizontal",
                 orientation: str = "down", bend_longitud: float = 0, bend_angle: float = 0, bend_height: float = 0,
                 transverse_center: tuple = None, denomination: None = str):

        super().__init__(width=0, height=reinforcement_length, type="bar", x=x, y=y)

        # Bending attributes.
        self.bend_longitud = bend_longitud
        self.bend_angle = bend_angle
        self.bend_height = bend_height
        self.bending_proyection = bend_height / tan(rads(bend_angle)) if bend_angle else 0

        # Mandrel attributes.
        self.left_anchor = left_anchor
        self.right_anchor = right_anchor
        self.mandrel_radius = mandrel_radius
        self.mandrel_radius_ext = diameter + mandrel_radius

        # Geometric attributes.
        self.diameter = diameter
        self.radius = diameter / 2
        self.direction = direction
        self.orientation = orientation
        self.transverse_center = transverse_center
        self.reinforcement_length = reinforcement_length
        self.length = (reinforcement_length + (1 / cos(rads(bend_angle)) - 1) * self.bending_proyection * 2
                       + left_anchor + right_anchor)
        self.weight = (diameter ** 2 * pi / 4) * self.length * STEEL_WEIGHT

        # Boxing attributes.
        self.box_width = reinforcement_length
        if self.left_anchor or right_anchor or bend_height:
            max_anchor = max([left_anchor, right_anchor])  # Maximum anchor.

            self.box_height = max([self.mandrel_radius_ext + max_anchor,
                                   diameter * 2 + self.bend_height])
        else:
            self.box_height = self.diameter

        # Others.
        self.denomination = denomination

    # Drawing longitudinal function.
    def draw_longitudinal(self, document: Drawing, x: float = None, y: float = None, unifilar: bool = False,
                          dimensions: bool = True, denomination: bool = True) -> list:
        if x is None:
            x = self.x
        if x is None:
            y = self.y

        diameter = self.diameter if not unifilar else 0
        mandrel_radius_ext = self.mandrel_radius_ext if not unifilar else 0
        sides_left_anchor = [0, 1, 1, 1] if not unifilar else [0, 0, 0, 1]
        sides_right_anchor = [0, 1, 1, 1] if not unifilar else [0, 1, 0, 0]

        # Configurations variables.
        text_dim_distance = 0.05
        text_dim_height = 0.05
        text_denom_distance = 0.05
        text_denom_height = 0.05

        # Drawing of simple entities and setting of variables for progressive draw.
        group = []  # Variable that cointains elements objetcs.
        group_text = []  # Variable that cointains text elements objetcs.

        length_first_rect_bar = self.reinforcement_length
        x_first_rect_bar = x
        sides_first_rect_bar = [1, 1, 1, 1]

        length_third_rect_bar = self.reinforcement_length
        sides_third_rect_bar = [1, 1, 1, 0]

        # Left anchor.
        if self.left_anchor:
            x_lab = x
            y_lab = y + self.box_height - mandrel_radius_ext - self.left_anchor
            center_point_lac = (x + mandrel_radius_ext,
                                y + self.box_height - mandrel_radius_ext)

            length_first_rect_bar -= mandrel_radius_ext
            x_first_rect_bar += mandrel_radius_ext
            sides_first_rect_bar[3] = 0

            # From left to right.
            # First anchor rect bar (left).
            group += rect(doc=document, width=diameter, height=self.left_anchor, x=x_lab, y=y_lab,
                          sides=sides_left_anchor)

            if not unifilar:
                # First bend curve.
                group += curve(doc=document, center_point=center_point_lac, radius=self.mandrel_radius,
                               start_angle=90, end_angle=180, thickness=diameter)

        # Right anchor.
        if self.right_anchor:
            x_rab = x + self.reinforcement_length - diameter
            y_rab = y + self.box_height - mandrel_radius_ext - self.right_anchor
            center_point_rac = (x + self.reinforcement_length - mandrel_radius_ext,
                                y + self.box_height - mandrel_radius_ext)

            length_first_rect_bar -= mandrel_radius_ext if not self.bend_longitud else 0
            sides_first_rect_bar[1] = 0
            length_third_rect_bar -= mandrel_radius_ext
            sides_third_rect_bar = [1, 0, 1, 0]

            # From left to right.
            # Sixth bend curve.
            if not unifilar:
                group += curve(doc=document, center_point=center_point_rac, radius=self.mandrel_radius, start_angle=0,
                               end_angle=90, thickness=diameter)

            # Second anchor rect bar (right).
            group += rect(doc=document, width=diameter, height=self.right_anchor, x=x_rab, y=y_rab,
                          sides=sides_right_anchor)

        # Bending bar.
        if self.bend_longitud:

            alpha = rads(self.bend_angle / 2)
            dx1 = cos(alpha) * sin(alpha) * diameter * 2
            dy1 = sin(alpha) * sin(alpha) * diameter * 2
            dx2 = cos(alpha) * sin(alpha) * diameter * 4
            dy2 = sin(alpha) * sin(alpha) * diameter * 4

            longitud_mid = (self.reinforcement_length + self.bend_longitud) / 2
            longitud_curves = dx1 + dx2
            longitud_proyeccion = (self.bend_height + diameter - dy1 - dy2) / tan(alpha * 2)

            length_first_rect_bar -= (longitud_mid + longitud_curves + longitud_proyeccion)
            sides_first_rect_bar[1] = 0
            length_third_rect_bar -= (longitud_mid + longitud_curves + longitud_proyeccion)

            center_point_bc_first = (x_first_rect_bar + length_first_rect_bar,
                                     y + self.box_height - diameter * 2)
            center_point_bc_second = (x + (self.reinforcement_length - self.bend_longitud) / 2,
                                      y + self.box_height - self.bend_height)

            # From left to right.
            # First piece of rect bar.
            sides_first_rect_bar = sides_first_rect_bar if not unifilar else [1, 0, 0, 0]
            group += rect(doc=document, width=length_first_rect_bar, height=diameter, x=x_first_rect_bar,
                          y=y + self.box_height - diameter, sides=sides_first_rect_bar)

            if not unifilar:
                # First curve of bend.
                group += curve(doc=document, center_point=center_point_bc_first, radius=diameter,
                               start_angle=90 - self.bend_angle, end_angle=90, thickness=diameter)

            # First bend rect bar.
            if not unifilar:
                group += line(doc=document,
                              p1=(x + (self.reinforcement_length - self.bend_longitud) / 2 - dx1,
                                  y + self.box_height - self.bend_height - diameter + dy1),
                              p2=(x + (self.reinforcement_length - self.bend_longitud) / 2 - dx1 - longitud_proyeccion,
                                  y + self.box_height - dy2))

            group += line(doc=document,
                          p1=(x + (self.reinforcement_length - self.bend_longitud) / 2 - dx2,
                              y + self.box_height - self.bend_height - diameter * 2 + dy2),
                          p2=(x + (self.reinforcement_length - self.bend_longitud) / 2 - dx2 - longitud_proyeccion,
                              y + self.box_height - diameter - dy1))

            # Second curve of bend.
            if not unifilar:
                group += curve(doc=document, center_point=center_point_bc_second, radius=diameter,
                               start_angle=270 - self.bend_angle, end_angle=270, thickness=diameter)

            # Second rect bar.
            sides_second_rect_bar = [1, 0, 1, 0] if not unifilar else [1, 0, 0, 0]
            group += rect(doc=document, width=self.bend_longitud, height=diameter,
                          x=x + (self.reinforcement_length - self.bend_longitud) / 2,
                          y=y + self.box_height - diameter * 2 - self.bend_height, sides=sides_second_rect_bar)

            # Third bend curve.
            if not unifilar:
                group += curve(doc=document, center_point=(x + longitud_mid,
                                                           y + self.box_height - self.bend_height),
                               radius=diameter, start_angle=270, end_angle=270 + self.bend_angle, thickness=diameter)

            # Second bend bar.
            if not unifilar:
                group += line(doc=document,
                              p1=(x + longitud_mid + dx1,
                                  y + self.box_height - self.bend_height - diameter + dy1),
                              p2=(x + longitud_mid + dx1 + longitud_proyeccion,
                                  y + self.box_height - dy2))

            group += line(doc=document,
                          p1=(x + longitud_mid + dx2,
                              y + self.box_height - self.bend_height - diameter * 2 + dy2),
                          p2=(x + longitud_mid + dx2 + longitud_proyeccion,
                              y + self.box_height - diameter - dy1))

            # Fourth bend curve.
            if not unifilar:
                group += curve(doc=document, center_point=(x + longitud_mid + longitud_curves + longitud_proyeccion,
                                                           y + self.box_height - diameter * 2),
                               radius=diameter, start_angle=90, end_angle=90 + self.bend_angle, thickness=diameter)

            # Third rect bar.
            sides_third_rect_bar = sides_third_rect_bar if not unifilar else [1, 0, 0, 0]
            group += rect(doc=document, width=length_third_rect_bar, height=diameter,
                          x=x + longitud_mid + longitud_curves + longitud_proyeccion,
                          y=y + self.box_height - diameter, sides=sides_third_rect_bar)

        else:
            # From left to right.
            # First rect bar (body).
            sides_first_rect_bar = sides_first_rect_bar if not unifilar else [1, 0, 0, 0]
            mandrel_radius_ext = 0 if not self.left_anchor and not self.right_anchor else mandrel_radius_ext
            group += rect(doc=document, width=length_first_rect_bar, height=diameter, x=x + mandrel_radius_ext,
                          y=y + self.box_height - diameter, sides=sides_first_rect_bar)

        if dimensions:
            if self.left_anchor:
                group += text(document=document, text="({:.2f})".format(self.left_anchor), height=text_dim_height,
                              point=(x_lab - text_dim_distance, y + self.left_anchor / 2), rotation=90,
                              attr={"halign": 4, "valign": 0})
                group_text += [group[-1]]

            if self.right_anchor:
                group += text(document=document, text="({:.2f})".format(self.right_anchor), height=text_dim_height,
                              point=(x_rab + text_dim_distance, y + self.right_anchor / 2), rotation=90,
                              attr={"halign": 4, "valign": 0})
                group_text += [group[-1]]

            group += text(document=document, text="({:.2f})".format(self.box_width), height=text_dim_height,
                          point=(x + self.box_width / 2, y + self.box_height - text_dim_distance), rotation=0,
                          attr={"halign": 4, "valign": 0})
            group_text += [group[-1]]

            if denomination:
                group += text(document=document, text=self.denomination, height=text_denom_height,
                              point=(x + self.box_width / 2,
                                     y + self.box_height / 2 + text_denom_height + text_denom_distance),
                              rotation=0,
                              attr={"halign": 4, "valign": 0})
                group_text += [group[-1]]

        if unifilar:
            translate(objects=group, vector=(0, -self.mandrel_radius_ext))

        # Orienting the bar (direction and orientation).
        self.__direc_orient(group, x=x, y=y, unifilar=unifilar)
        if self.orientation == "top":
            self.__direct_orient_text(group_text)

        return group

    # Drawing of transverse section of bar function.
    def draw_transverse(self, document: Drawing, x: float = None, y: float = None):

        if x is None:
            x = self.x

        if y is None:
            y = self.y

        # Checking if there is a defined transverse center.
        if self.transverse_center:
            x += self.transverse_center[0]
            y += self.transverse_center[1]

        # Drawing of circle.
        group = circle(doc=document, center_point=(x + self.radius, y + self.radius), radius=self.radius)

        return group

    # Function that orients drawing.
    def __direc_orient(self, group: list, x: float = None, y: float = None, unifilar: bool = False):
        if x is None:
            x = self.x

        if y is None:
            y = self.y

        # Orienting the bar (direction and orientation).
        # Direction.
        if self.direction == "vertical":
            pivot_point = (x, y + self.box_height)
            vector_translate = (x - (pivot_point[0] * cos(pi / 2) - pivot_point[1] * sin(pi / 2)),
                                y - (pivot_point[0] * sin(pi / 2) + pivot_point[1] * cos(pi / 2)))

            rotate(group, pi / 2)
            translate(group, vector=vector_translate)

        # Orientation.
        if self.orientation == "top":
            mirror(objects=group, mirror_type="x")
            if self.direction == "vertical":
                translate(group, vector=(0, self.box_width + y * 2))
            else:
                translate(group, vector=(0, self.box_height + y * 2))
                if unifilar:
                    translate(group, vector=(0, self.radius))

        elif self.orientation == "right":
            if self.direction == "vertical":
                if unifilar:
                    translate(group, vector=(-self.mandrel_radius_ext, 0))
            else:
                pass

        elif self.orientation == "left":
            mirror(objects=group, mirror_type="y")

            if self.direction == "vertical":
                translate(group, vector=(self.box_height + x * 2, 0))
            else:
                translate(group, vector=(self.box_width + x * 2, 0))

    def __direct_orient_text(self, group: list, x: float = None, y: float = None):
        coordinates = [entitie.get_placement()[1] for entitie in group]
        mirror(group, mirror_type="x")

        for i, entitie in enumerate(group):
            entitie.set_placement([-coordinates[i][0], coordinates[i][1], coordinates[i][2]])



# Imports.
# Local imports.

# External imports.


# Function that returns list of bar numbers.
def expand_dictionary(dictionary) -> list:
    expanded_list = []
    for key, value in dictionary.items():

        expanded_list.extend([key] * value)

    return expanded_list


# Function that returns a simetric list.
def gen_symmetric_list(dictionary: dict, nomenclature: str = None, number_init: int = None, factor: float = 1) -> tuple:

    if nomenclature is None:
        nomenclature = "#"

    if number_init is None:
        n = 0
    else:
        n = number_init

    first_odd = False
    symmetryc_list = []
    denomination_list = []

    for key, value in dictionary.items():
        key_factored = key / factor
        for i in range(value // 2):
            symmetryc_list.insert(0, key_factored)
            symmetryc_list.insert(len(symmetryc_list), key_factored)
            denomination_list.insert(0, "{2}{3} {0}Ø{1}".format(value, key, nomenclature, n))
            denomination_list.insert(len(denomination_list), "{2}{3} {0}Ø{1}".format(value, key, nomenclature, n))

        if is_odd(value):
            symmetryc_list.insert(len(symmetryc_list) // 2, key_factored)
            denomination_list.insert(len(denomination_list) // 2, "{2}{3} {0}Ø{1}".format(value, key, nomenclature, n))
            if first_odd:
                return ()
            else:
                first_odd = True
        n += 1

    return symmetryc_list, denomination_list


# Function that check is an odd number.
def is_odd(number):
    return bool(number % 2)


# Function that analize str for bars.
def str_to_dict_bar(data: str) -> dict:
    data_list = data.replace(" ", "").split("+")

    data_dict = {}
    for element in data_list:
        key_value = element.split("db")

        data_dict[int(key_value[1])] = int(key_value[0])

    data_dict_ordered = {k: v for k, v in sorted(data_dict.items(), key=lambda item: item[0], reverse=False)}

    return data_dict_ordered



# Imports.
# Local imports.
from drawing_utils import delimit_axe, dim_linear, rect, text
from bar import Bar
from element import Element
from stirrup import Stirrup
from utils import gen_symmetric_list, str_to_dict_bar

# External imports.
import ezdxf
from ezdxf.document import Drawing
from ezdxf.gfxattribs import GfxAttribs
from ezdxf import zoom


class Beam(Element):
    def __init__(self, width: float, height: float, length: float, as_sup: dict = None, as_right: dict = None,
                 as_inf: dict = None, as_left: dict = None, anchor_sup: float | list = 0, anchor_right: float | list = 0,
                 anchor_inf: float | list = 0, anchor_left: float | list = 0, stirrups_db: list = None,
                 stirrups_sep: list = None, stirrups_length: list = None, stirrups_anchor: list = None,
                 stirrups_x: list = None, columns: list = None, columns_pos: list = None, columns_symbol: list = None,
                 cover: float = 0.025, x: float = 0, y: float = 0, direction: str = "horizontal",
                 orientation: str = "right", nomenclature: str = None, number_init: int = None):

        super().__init__(width=width, height=height, type="beam", x=x, y=y)

        # Armor attributes.
        if as_sup:
            self.as_sup = as_sup
            self.max_db_sup = max(self.as_sup.keys())
            self.anchor_sup = anchor_sup if type(anchor_sup) == list else [anchor_sup] * sum(self.as_sup.values())
            self.number_init_sup = number_init if number_init is not None else 0
            number_init = number_init + len(as_sup) if number_init else len(as_sup)
        else:
            self.as_sup, self.max_db_sup, self.anchor_sup = {}, 0, 0

        if as_right:
            self.as_right = as_right
            self.max_db_right = max(self.as_right.keys())
            self.anchor_right = anchor_right if type(anchor_right) == list else [anchor_right] * sum(self.as_right.values())
            self.number_init_right = number_init
            number_init += len(as_right)
        else:
            self.as_right, self.max_db_right, self.anchor_right = {}, 0, 0

        if as_inf:
            self.as_inf = as_inf
            self.max_db_inf = max(self.as_inf.keys())
            self.anchor_inf = anchor_inf if type(anchor_inf) == list else [anchor_inf] * sum(self.as_inf.values())
            self.number_init_inf = number_init
            number_init += len(as_inf)
        else:
            self.as_inf, self.max_db_inf, self.anchor_inf = {}, 0, 0

        if as_left:
            self.as_left = as_left
            self.max_db_left = max(self.as_left.keys()) if self.as_right else 0
            self.anchor_left = anchor_left if type(anchor_left) == list else [anchor_left] * sum(self.as_left.values())
            self.number_init_left = number_init
        else:
            self.as_left, self.max_db_left, self.anchor_left = {}, 0, 0

        self.cover = cover

        # Stirrups attributes.
        self.stirrups_db = stirrups_db
        self.stirrups_sep = stirrups_sep
        self.stirrups_length = stirrups_length

        if stirrups_anchor is None:
            self.stirrups_anchor = [0.1] * len(stirrups_db)
        else:
            self.stirrups_anchor = stirrups_anchor

        self.stirrups_x = stirrups_x

        # Geometric attributes.
        self.length = length
        self.direction = direction
        self.orientation = orientation

        # Columns attributes.
        self.columns = columns
        self.columns_pos = columns_pos

        if columns_symbol is None:
            self.columns_symbol = range(1, len(self.columns) + 1)
        else:
            self.columns_symbol = columns_symbol

        # Elements/Entities.
        self.bars_as_sup = self.__dict_to_bars(self.as_sup, width=self.width, x=x, y=y, side=0,
                                               anchor=self.anchor_sup, nomenclature=nomenclature,
                                               number_init=self.number_init_sup) if as_sup else []
        self.bars_as_right = self.__dict_to_bars(self.as_right, width=height, x=x, y=y, side=1,
                                                 anchor=self.anchor_right, nomenclature=nomenclature,
                                                 number_init=self.number_init_right) if as_right else []
        self.bars_as_inf = self.__dict_to_bars(self.as_inf, width=self.width, x=x, y=y, side=2,
                                               anchor=self.anchor_inf, nomenclature=nomenclature,
                                               number_init=self.number_init_inf) if as_inf else []
        self.bars_as_left = self.__dict_to_bars(self.as_left, width=height, x=x, y=y, side=3,
                                                anchor=self.anchor_left, nomenclature=nomenclature,
                                                number_init=self.number_init_left) if as_left else []
        self.all_bars = self.bars_as_sup + self.bars_as_right + self.bars_as_inf + self.bars_as_left

        self.stirrups = self.__list_to_stirrups()

        self.all_elements = self.all_bars + self.stirrups

    # Function that draws beam along longitudinal axe.
    def draw_longitudinal(self, document: Drawing, x: float = None, y: float = None, concrete_shape: bool = True,
                          bars: bool = True, columns: bool = True, columns_axes: bool = True,
                          stirrups: bool = True, middle_axe: bool = True, middle_axe_symbol: str = "A",
                          dim: bool = True, dim_style: str = "EZ_M_25_H25_CM", unifilar_bars: bool = False,
                          unifilar_stirrups: bool = True) -> list:

        if x is None:
            x = self.x
        if y is None:
            y = self.y

        group = []
        delimit_axe_height = self.height * 4
        delimit_axe_y = y - self.height

        # Drawing concrete shape.
        if concrete_shape:
            group += rect(doc=document, width=self.length, height=self.height, x=x, y=y, sides=[1, 1, 1, 1],
                          polyline=True)  # Body.

        # Drawing columns.
        for i, x_column in enumerate(self.columns_pos):
            x_column_relative = x + x_column
            x_delimit_axe_relative = x_column_relative + self.columns[i][0] / 2
            if columns:
                group += rect(doc=document, width=self.columns[i][0], height=self.columns[i][1], x=x_column_relative,
                              y=y, fill=True, polyline=True)
            if columns_axes:
                group += delimit_axe(document=document, x=x_delimit_axe_relative, y=delimit_axe_y,
                                     height=delimit_axe_height, radius=0.1, text_height=0.1,
                                     symbol=self.columns_symbol[i], attr=GfxAttribs(linetype="CENTER"))

        # Drawing axe.
        if middle_axe:
            group += delimit_axe(document=document, x=x - self.height, y=y + self.height / 2,
                                 height=self.length + self.height * 2, symbol=middle_axe_symbol, direction=0,
                                 attr=GfxAttribs(linetype="CENTER"))

        # Drawing of stirrups.
        if stirrups:
            for stirrup in self.stirrups:
                group += stirrup.draw_longitudinal(document=document, x=x + stirrup.x, y=y + stirrup.y,
                                                   unifilar=unifilar_stirrups)

        # Drawing of bars.
        if bars:
            self.bars_as_sup[0].draw_longitudinal(document=document,
                                                  x=x + self.bars_as_sup[0].x,
                                                  y=y + self.bars_as_sup[0].y,
                                                  unifilar=unifilar_bars,
                                                  dimensions=False)  # Only mayor bar.
            self.bars_as_inf[0].draw_longitudinal(document=document,
                                                  x=x + self.bars_as_inf[0].x,
                                                  y=y + self.bars_as_inf[0].y,
                                                  unifilar=unifilar_bars,
                                                  dimensions=False)  # Only inferior mayor bar.
            for bar in self.bars_as_left:
                bar.draw_longitudinal(document=document,
                                      x=x + bar.x,
                                      y=y + bar.y,
                                      unifilar=unifilar_bars,
                                      dimensions=False)  # Only left bars.

        # Drawing dimensions.
        if dim:
            dim_y = y + self.height * 2
            dim_p_base = (x + self.length / 2, y + delimit_axe_height * 0.65)

            for stirrup in self.stirrups:
                group += dim_linear(document=document,
                                    p_base=(x + stirrup.x + stirrup.reinforcement_length / 2, dim_y),
                                    p1=(x + stirrup.x, dim_y),
                                    p2=(x + stirrup.x + stirrup.reinforcement_length, dim_y),
                                    dimstyle=dim_style)

            group += dim_linear(document=document,
                                p_base=dim_p_base,
                                p1=(x, y + self.height),
                                p2=(x + self.length, y + self.height),
                                dimstyle=dim_style)

        return group

    # Function that draws transverse section.
    def draw_transverse_rebar_detailing(self, document: Drawing, x: float = None, y: float = None,
                                        x_section: float = None, unifilar: bool = False,
                                        dimensions: bool = True) -> list:

        stirrups = self.__elements_section(elements=self.stirrups, x=x_section)

        group = []
        for stirrup in stirrups:
            group += stirrup.draw_transverse(document=document, x=x, y=y, unifilar=unifilar, dimensions=dimensions)

        return group

    def draw_transverse(self, document: Drawing, x: float, y: float, x_section: float = None,
                        unifilar: bool = False, dimensions: bool = True) -> list:

        if x_section is None:
            x_section = self.length / 2

        # Checking if x given is in beam length.
        if not self.x <= x_section <= self.x + self.length:
            return []

        # Configurations variables.
        text_dim_distance = 0.05
        text_dim_height = 0.05

        # Obtaining elements for given x_section.
        entities = self.__elements_section(x=x_section)

        # Filtering elements.
        bars = filter(lambda e: e.type == "bar", entities)
        stirrups = filter(lambda e: e.type == "stirrup", entities)

        # Concrete shape drawing.
        group = rect(doc=document, width=self.width, height=self.height, x=x, y=y, polyline=True)

        # Drawing of bars.
        for bar in bars:
            group += bar.draw_transverse(document=document, x=x, y=y)

        # Drawing of stirrups.
        for stirrup in stirrups:
            delta_x = self.cover - max(self.max_db_sup, self.max_db_inf) / 2000 - stirrup.diameter

            group += stirrup.draw_transverse(document=document, x=x + delta_x, y=y + stirrup.y,
                                             unifilar=unifilar)

        # Drawing of dimensions.
        if dimensions:
            group += dim_linear(document=document, p_base=(x - text_dim_distance, y + self.height / 2),
                                p1=(x, y), p2=(x, y + self.height), rotation=90, dimstyle="EZ_M_10_H25_CM")
            group += dim_linear(document=document, p_base=(x + self.width / 2, y + self.height + text_dim_distance),
                                p1=(x, y + self.height), p2=(x + self.width, y + self.height),
                                dimstyle="EZ_M_10_H25_CM")

        return group

    # Function that draws the rebar detailing.
    def draw_longitudinal_rebar_detailing(self, document: Drawing, x: float = None, y: float = None,
                                          unifilar: bool = True, columns_axes: bool = True) -> list:

        def __draw_section_bars(bars: list = None, spacing: float = 0.2, rebar_x: float = None,
                                rebar_y: float = None, barline=True, text_reference: str = None) -> tuple:
            text_height = 0.05
            entities = []
            graph_list = []

            if text_reference:
                entities += text(document=document, text=text_reference, height=text_height,
                                 point=(rebar_x - self.height * 2 * 0.8, rebar_y + self.height * 0.05 - spacing))

            for bar in bars:
                if bar.denomination not in graph_list:
                    rebar_y -= spacing + bar.box_height
                    entities += bar.draw_longitudinal(document=document, x=rebar_x, y=rebar_y, unifilar=unifilar,
                                                      dimensions=True)
                    graph_list.append(bar.denomination)
            rebar_y -= spacing
            if barline:
                entities += rect(doc=document, width=self.length + self.height * 4, height=0,
                                 x=rebar_x - self.height * 2, y=rebar_y, sides=(1, 0, 0, 0))

            return entities, rebar_x, rebar_y

        group, rebar_x, rebar_y = [], x, y
        if self.as_sup:
            barline = True if self.as_right or self.bars_as_left or self.bars_as_inf else False
            entities, rebar_x, rebar_y = __draw_section_bars(bars=self.bars_as_sup, spacing=0.2, rebar_x=rebar_x,
                                                             rebar_y=rebar_y, barline=barline, text_reference="As sup.")
            group += entities

        if self.as_left:
            barline = True if self.bars_as_left or self.bars_as_inf else False
            entities, rebar_x, rebar_y = __draw_section_bars(bars=self.bars_as_right, spacing=0.2, rebar_x=rebar_x,
                                                             rebar_y=rebar_y, barline=barline, text_reference="As right.")
            group += entities

        if self.as_left:
            barline = True if self.bars_as_inf else False
            entities, rebar_x, rebar_y = __draw_section_bars(bars=self.bars_as_left, spacing=0.2, rebar_x=rebar_x,
                                                             rebar_y=rebar_y, barline=barline, text_reference="As left.")
            group += entities

        if self.as_inf:
            entities, rebar_x, rebar_y = __draw_section_bars(bars=self.bars_as_inf, spacing=0.2, rebar_x=rebar_x,
                                                             rebar_y=rebar_y, barline=False, text_reference="As inf.")
            group += entities

        if columns_axes:
            for i, x_column in enumerate(self.columns_pos):
                delimit_axe_height = y - rebar_y
                delimit_axe_y = rebar_y
                x_column_relative = x + x_column
                x_delimit_axe_relative = x_column_relative + self.columns[i][0] / 2

                group += delimit_axe(document=document, x=x_delimit_axe_relative, y=delimit_axe_y,
                                     height=delimit_axe_height, radius=0.1, text_height=0.1,
                                     attr=GfxAttribs(linetype="CENTER"))

        return group

    def draw_table_rebar_detailing(self, x: float = None, y: float = None, unifilar: bool = False):
        pass

    def __dict_to_bars(self, bars: dict, width: float, x: float, y: float, side: int, anchor: list,
                       nomenclature: str = None, number_init: int = None) -> list:

        if side < 0 or side > 3:
            raise ValueError

        list_bars, list_denom = gen_symmetric_list(dictionary=bars, nomenclature=nomenclature, number_init=number_init,
                                                   factor=1000)  # Creating symetric list of bars.

        # Calculating separtion and definition of initial x_sep and y_sep.
        if side == 0 or side == 2:  # Sides top (0) or bottom (2).
            db_max = max(self.max_db_sup, self.max_db_inf) / 1000
            width_util = width - self.cover * 2 + (db_max - max(list_bars))
            separation = width_util / (len(list_bars) - 1)
            x_sep, y_sep = -separation, 0

        else:  # Sides right (1) or left (3).
            db_max = max([self.max_db_right, self.max_db_left]) / 1000
            width_util = width - self.cover * 2
            separation = width_util / (len(list_bars) + 1)
            x_sep, y_sep = 0, 0

        # Generating bars.
        entities, delta_x, delta_y, orientation = [], 0, 0, "bottom"
        for i, db in enumerate(list_bars):
            # Calculation of x_delta and y_delta.
            if side == 0 or side == 2:
                x_sep += separation
                delta_x = -self.cover + db_max / 2
                if side == 0:
                    delta_y = self.height - self.cover - 3 * db / 2 - anchor[i]
                    delta_y_transverse = self.height - self.cover + self.max_db_sup / 2000 - db
                    orientation = "right"
                if side == 2:
                    delta_y = self.cover - self.max_db_inf / 2000
                    delta_y_transverse = self.cover - self.max_db_inf / 2000
                    orientation = "top"
            else:
                y_sep += separation
                delta_y = self.cover - db / 2

                if side == 1:
                    delta_x = self.cover + db - max([self.max_db_sup, self.max_db_inf, self.max_db_right]) / 2000 - self.width
                    delta_y_transverse = delta_y
                    orientation = "right"
                if side == 3:
                    delta_x = - self.cover + max([self.max_db_sup, self.max_db_inf, self.max_db_left]) / 2000
                    delta_y_transverse = delta_y
                    orientation = "right"

            x_bar_long = x + self.cover
            x_bar_transverse = x_sep - delta_x
            y_bar_long = y + y_sep + delta_y
            y_bar_transverse = y_sep + delta_y_transverse

            entities.append(Bar(reinforcement_length=self.length - self.cover * 2, diameter=db,
                                x=x_bar_long, y=y_bar_long,
                                left_anchor=anchor[i], right_anchor=anchor[i], mandrel_radius=db,
                                orientation=orientation,
                                transverse_center=(x_bar_transverse, y_bar_transverse), denomination=list_denom[i]))

        return entities

    # Function that returns elements in a given x_beam position.
    def __elements_section(self, elements: list = None, x: float = None) -> list:
        if elements is None:
            elements = self.all_elements

        if x is None:
            x = self.length / 2  # Default value if is nothing entered.

        # Filtering elements by X coordinate.
        entities = []
        for element in elements:
            if element.x <= x <= element.x + element.reinforcement_length:
                entities.append(element)

        return entities

    # Function that creates stirrups from list.
    def __list_to_stirrups(self) -> list:

        stirrup_width = self.width - (self.cover - max(self.max_db_sup, self.max_db_inf) / 2000) * 2
        stirrup_height = self.height - (self.cover * 2 - self.max_db_sup / 2000 - self.max_db_inf / 2000)
        stirrups = zip(self.stirrups_db, self.stirrups_length, self.stirrups_sep, self.stirrups_anchor, self.stirrups_x)
        mandrel_radius_sup = self.max_db_sup / 2000
        mandrel_radius_inf = self.max_db_inf / 2000

        # Generating stirrups.
        entities = []  # Empty list that will contain stirrups elements.
        for stirrup in stirrups:
            stirrup_x = self.x + stirrup[4]  # Difining x coordinate.
            stirrup_y = self.cover - self.max_db_inf / 2000 - stirrup[0]  # Subract of thick of stirrup.

            # Creating of stirrup.
            entities.append(Stirrup(width=stirrup_width + stirrup[0] * 2, height=stirrup_height + stirrup[0] * 2,
                                    diameter=stirrup[0], reinforcement_length=stirrup[1], spacing=stirrup[2],
                                    x=stirrup_x, y=stirrup_y, mandrel_radius_top=mandrel_radius_sup,
                                    mandrel_radius_bottom=mandrel_radius_inf, anchor=stirrup[3]))

        return entities



# Imports.
# Local imports.

from drawing_utils import (curve, dim_linear, get_lines_intersec, line, mirror, rads, rect_border_curve, rotate,
                           text, translate)
from element import Element
from globals import STEEL_WEIGHT

# External imports.
import ezdxf
from ezdxf.document import Drawing
from math import cos, sin, pi, floor

# Constants.
SIN45, COS45 = sin(rads(45)), cos(rads(45))


class Stirrup(Element):
    """
    Stirrup element, computes geometrics and physics props and manages dxf drawing methods.

        :param width: External width of stirrup.
        :type width: float
        :param height: External height of stirrup.
        :type height: float
        :param diameter: Diameter of stirrup bar.
        :type diameter: float
        :param reinforcement_length: Length of the stirrup reinforcement.
        :type reinforcement_length: float
        :param spacing: Stirrup spacing.
        :type spacing: float
        :param x: X coordinate of the bottom left corner Stirrup bounding box point.
        :type x: float
        :param y: Y coordinate of the bottom left corner Stirrup bounding box point.
        :type y: float
        :param mandrel_radius_top: Mandrel radius of stirrup bar at the top.
        :type mandrel_radius_top: float
        :param mandrel_radius_bottom: Mandrel radius of stirrup bar at the bottom.
        :type mandrel_radius_bottom: float
        :param anchor: Anchor length of stirrup.
        :type anchor: float
        :param direction: Direction of the stirrup (Horizontal or Vertical).
        :type direction: str
        :param orientation: Orientation of the stirrup (top, right, down, left).
        :type orientation: str

        :ivar reinforcement_length: Length of the stirrup reinforcement.
        :ivar spacing: Stirrup spacing.
        :ivar mandrel_radius_top: Mandrel radius of stirrup bar at the top.
        :ivar mandrel_radius_ext_top: Mandrel radius plus diameter of stirrup bar at the top.
        :ivar mandrel_radius_bottom: Mandrel radius of stirrup bar at the bottom.
        :ivar mandrel_radius_ext_bottom: Mandrel radius plus diameter of stirrup bar at the bottom.
        :ivar anchor: Anchor length of stirrup.
        :ivar quantity: Number of stirrups that forms de reinforcement.
        :ivar diameter: Diameter of stirrup bar.
        :ivar direction: Direction of the stirrup (Horizontal or Vertical).
        :ivar orientation: Orientation of the Stirrup (top, right, down, left).
        :ivar length: Length of the bar to fork the stirrup.
        :ivar weight: Weight of the stirrup, considering 7850 kg / m3.
        :ivar box_width: Width of the box that contains the stirrup.
        :ivar box_height: Height of the box that contains the stirrup.
    """
    def __init__(self, width: float, height: float, diameter: float, reinforcement_length: float,
                 spacing: float, x: float = 0, y: float = 0, mandrel_radius_top: float = 0,
                 mandrel_radius_bottom: float = 0, anchor: float = 0.15, direction: str = "horizontal",
                 orientation: str = "down"):
        """
        Initialeze a new instance of Stirrup.
        """
        super().__init__(width=width, height=height, type="stirrup", x=x, y=y)

        # Stirrups attributes.
        self.reinforcement_length = reinforcement_length
        self.spacing = spacing
        self.mandrel_radius_top = mandrel_radius_top
        self.mandrel_radius_ext_top = mandrel_radius_top + diameter
        self.mandrel_radius_bottom = mandrel_radius_bottom
        self.mandrel_radius_ext_bottom = mandrel_radius_bottom + diameter
        self.anchor = anchor
        self.quantity = floor((reinforcement_length * 100) / (spacing * 100)) + 1

        # Geometric attributes.
        self.diameter = diameter
        self.direction = direction
        self.orientation = orientation
        self.length = (width + height + anchor) * 2
        self.weight = (diameter ** 2 * pi / 4) * self.length * STEEL_WEIGHT

        # Box attributes.
        self.box_width = self.width
        self.box_height = self.height

    # Drawing longitudinal function.
    def draw_longitudinal(self, document: Drawing, x: float = None, y: float = None, unifilar=True):
        """
        Draw the longitudinal reinforcement of the stirrup in the dxf file.

        :param document: Document in which it will be drawn.
        :type document: Drawing
        :param x: X coordinate of the bottom corner of the drawing.
        :type x: float
        :param y: Y coordinate of the bottom corner of the drawing.
        :type y: float
        :param unifilar: Single-line drawing.
        :type unifilar: bool
        :return: None.
        :rtype: None
        """
        if x is None:
            x = self.x

        if y is None:
            y = self.y

        if unifilar:
            group = [line(doc=document, p1=(x + self.spacing * i, y), p2=(x + self.spacing * i, y + self.height))[0]
                     for i in range(self.quantity)]
        else:
            group = [line(doc=document, p1=(x + self.spacing * i, y), p2=(x + self.spacing * i, y + self.height))[0]
                     for i in range(self.quantity)]

        # Orienting the bar (direction and orientation).
        self.__direc_orient(group, x=x, y=y, longitudinal=True)

        return group

    # Drawing transverse section of stirrup function.
    def draw_transverse(self, document: Drawing, x: float = None, y: float = None, unifilar: bool = False,
                        dimensions: bool = False) -> list:
        """
        Draw the cross-section of the stirrup in the dxf file.

        :param document: Document in which it will be drawn.
        :type document: Drawing
        :param x: X coordinate of the bottom corner of the drawing.
        :type x: float
        :param y: Y coordinate of the bottom corner of the drawing.
        :type y: float
        :param unifilar: Single-line drawing.
        :type unifilar: bool
        :param dimensions: Dimensions drawing.
        :type dimensions: bool
        :return: None.
        :rtype: None
        """
        if x is None:
            x = self.x

        if y is None:
            y = self.y

        # Configurations variables.
        text_dim_distance = 0.05
        text_dim_height = 0.05

        # Curves radius of rect border bordes.
        curves_radius = [self.mandrel_radius_top, self.mandrel_radius_top, self.mandrel_radius_bottom,
                         self.mandrel_radius_bottom]

        # Constants equation of anchor lines.
        diff = 0 if unifilar else self.diameter
        m_45 = -1
        b_top_anchor_int = ((y + self.height - self.mandrel_radius_top - diff + self.mandrel_radius_top * SIN45)
                            - (m_45 * (x + self.mandrel_radius_top + diff + self.mandrel_radius_top * COS45)))
        b_top_anchor_ext = ((y + self.height - self.mandrel_radius_top - diff + (self.mandrel_radius_top + diff) * SIN45)
                            - (m_45 * (x + self.mandrel_radius_top + diff + (self.mandrel_radius_top + diff) * COS45)))
        b_bottom_anchor_int = ((y + self.height - self.mandrel_radius_top - diff - self.mandrel_radius_top * SIN45)
                               - (m_45 * (x + self.mandrel_radius_top + diff - self.mandrel_radius_top * COS45)))
        b_bottom_anchor_ext = ((y + self.height - self.mandrel_radius_top - diff - (self.mandrel_radius_top + diff) * SIN45)
                               - (m_45 * (x + self.mandrel_radius_top + diff - (self.mandrel_radius_top + diff) * COS45)))

        # Constants equation calculations of sides (top/left) lines.
        m_top_side = 0
        b_top_side = y + self.height if unifilar else y + self.height - self.diameter
        m_left_side = 9999999999
        b_left_side = -x * m_left_side if unifilar else -(x + self.diameter) * m_left_side

        # Constants equation of 45° circle line.
        m_45_circle_line = 1
        if not unifilar:
            b_45_circle_line = ((y + self.box_height - self.mandrel_radius_ext_top)
                                - m_45_circle_line * (x + self.mandrel_radius_ext_top))
        else:
            b_45_circle_line = ((y + self.box_height - self.mandrel_radius_top)
                                - m_45_circle_line * (x + self.mandrel_radius_top))

        # Calculations.
        intersec_top_anchor_int = get_lines_intersec(m_45, b_top_anchor_int, m_45_circle_line, b_45_circle_line)
        intersec_top_anchor_ext = get_lines_intersec(m_45, b_top_anchor_ext, m_top_side, b_top_side)
        intersect_bottom_anchor_int = get_lines_intersec(m_45, b_bottom_anchor_int, m_45_circle_line, b_45_circle_line)
        intersect_bottom_anchor_ext = get_lines_intersec(m_45, b_bottom_anchor_ext, m_45_circle_line, b_45_circle_line)

        if unifilar:
            # Calculations.
            p1_top_anchor_int = (intersec_top_anchor_int,
                                 intersec_top_anchor_int * m_45 + b_top_anchor_int)
            p2_top_anchor_int = (p1_top_anchor_int[0] + COS45 * self.anchor,
                                 p1_top_anchor_int[1] - SIN45 * self.anchor)

            p1_bottom_anchor_int = (intersect_bottom_anchor_int,
                                    intersect_bottom_anchor_int * m_45 + b_bottom_anchor_int)
            p2_bottom_anchor_int = (p1_bottom_anchor_int[0] + COS45 * self.anchor,
                                    p1_bottom_anchor_int[1] - SIN45 * self.anchor)

            p1_bottom_anchor_ext = p1_bottom_anchor_int
            p2_bottom_anchor_ext = p2_bottom_anchor_int

            # Anchor drawing.
            group = line(doc=document, p1=p1_top_anchor_int, p2=p2_top_anchor_int)  # Internal line top anchor.
            group += line(doc=document, p1=p1_bottom_anchor_int, p2=p2_bottom_anchor_int)  # Internal line bot anchor.

            # Rectangle drawing (circle borders).
            group += rect_border_curve(doc=document, width=self.width, height=self.height,
                                       radius=self.mandrel_radius_top, x=x, y=y, thickness=0, sides=[1, 1, 1, 1],
                                       curves_radius=curves_radius)

        else:
            # Drawing of side borders.
            group = rect_border_curve(doc=document, width=self.width - 2 * self.diameter,
                                      height=self.height - 2 * self.diameter, radius=self.mandrel_radius_top,
                                      x=x + self.diameter, y=y + self.diameter, thickness=self.diameter,
                                      sides=[1, 1, 1, 0], curves=[0, 1, 1, 1], curves_radius=curves_radius)

            # External left side border.
            group += line(doc=document, p1=(x, y + self.mandrel_radius_ext_bottom),
                          p2=(x, y + self.box_height - self.mandrel_radius_ext_top))

            # Drawing of anchors.
            # Calculatios of points.
            p1_top_anchor_int = (intersec_top_anchor_int,
                                 intersec_top_anchor_int * m_45 + b_top_anchor_int)
            p2_top_anchor_int = (p1_top_anchor_int[0] + COS45 * self.anchor,
                                 p1_top_anchor_int[1] - SIN45 * self.anchor)
            p1_top_anchor_ext = (intersec_top_anchor_ext,
                                 intersec_top_anchor_ext * m_45 + b_top_anchor_ext)
            p2_top_anchor_ext = (p2_top_anchor_int[0] + self.diameter * COS45,
                                 p2_top_anchor_int[1] + self.diameter * SIN45)
            p1_bottom_anchor_int = (intersect_bottom_anchor_int,
                                    m_45 * intersect_bottom_anchor_int + b_bottom_anchor_int)
            p2_bottom_anchor_int = (p1_bottom_anchor_int[0] + COS45 * self.anchor,
                                    p1_bottom_anchor_int[1] - SIN45 * self.anchor)
            p1_bottom_anchor_ext = (intersect_bottom_anchor_ext,
                                    m_45 * intersect_bottom_anchor_ext + b_bottom_anchor_ext)
            p2_bottom_anchor_ext = (p1_bottom_anchor_ext[0] + COS45 * self.anchor,
                                    p1_bottom_anchor_ext[1] - SIN45 * self.anchor)

            # Center of top left curve.
            center_point_curve = (x + self.mandrel_radius_ext_top, y + self.box_height - self.mandrel_radius_ext_top)

            # Drawing of lines anchors.
            group += line(doc=document, p1=p1_top_anchor_ext, p2=p2_top_anchor_ext)
            group += line(doc=document, p1=p1_top_anchor_int, p2=p2_top_anchor_int)
            group += line(doc=document, p1=p1_bottom_anchor_int, p2=p2_bottom_anchor_int)
            group += line(doc=document, p1=p1_bottom_anchor_ext, p2=p2_bottom_anchor_ext)

            group += line(doc=document, p1=p2_bottom_anchor_int, p2=p2_bottom_anchor_ext)  # Closed lines bottom.
            group += line(doc=document, p1=p2_top_anchor_int, p2=p2_top_anchor_ext)  # Closed lines top.

            # Left internal side.
            group += line(doc=document, p1=(x + self.diameter, y + self.mandrel_radius_ext_bottom),
                          p2=(p1_top_anchor_ext[0] - (self.diameter * 2 + self.mandrel_radius_top * 2) * COS45,
                              p1_top_anchor_ext[1] - (self.diameter * 2 + self.mandrel_radius_top * 2) * SIN45))

            # Top left curve.
            group += curve(doc=document, center_point=center_point_curve, start_angle=90, end_angle=225,
                           radius=self.mandrel_radius_top + self.diameter, thickness=0)  # Internal.
            group += curve(doc=document, center_point=center_point_curve, start_angle=45, end_angle=225,
                           radius=self.mandrel_radius_top, thickness=0)  # External.

        if dimensions:
            group += dim_linear(document=document, p_base=(x - text_dim_distance, y + self.height / 2),
                                p1=(x, y), p2=(x, y + self.height), rotation=90,
                                dimstyle="EZ_M_10_H25_CM")
            group += dim_linear(document=document,
                                p_base=(x + self.width / 2 - text_dim_distance * 2.5,
                                        y + self.height - text_dim_distance * 2.5),
                                p1=p1_bottom_anchor_ext, p2=p2_bottom_anchor_ext, rotation=315,
                                dimstyle="EZ_M_10_H25_CM")
            group += dim_linear(document=document, p_base=(x + self.width / 2, y + self.height + text_dim_distance),
                                p1=(x, y + self.height), p2=(x + self.width, y + self.height),
                                dimstyle="EZ_M_10_H25_CM")
            group += text(document=document, text="L: {0:.2f}".format(self.length), height=text_dim_height,
                          point=(x + self.width / 2, y - 2 * text_dim_distance), attr={"halign": 4, "valign": 0})

        # Orienting the bar (direction and orientation).
        self.__direc_orient(group, x=x, y=y, longitudinal=False)

        return group

    # Function that orients drawing.
    def __direc_orient(self, group: list, x: float = None, y: float = None, longitudinal: bool = False):
        """
        Orient the elements according to the parameters of the stirrup in the group entered.

        :param group: Group of elements dxf to orient.
        :param x: X coordinate of the bottom corner of the drawing.
        :type x: float
        :param y: Y coordinate of the bottom corner of the drawing.
        :type y: float
        :param longitudinal: Type of drawing to which the dxf elements belong (longitudinal or transverse).
        :type longitudinal: bool
        :return: Group of elements dxf to oriented.
        :rtype: list
        """
        # Orienting the bar (direction and orientation).
        # Direction.
        if self.direction == "vertical":
            pivot_point = (x, y + self.box_height)
            vector_translate = (x - (pivot_point[0] * cos(pi / 2) - pivot_point[1] * sin(pi / 2)),
                                y - (pivot_point[0] * sin(pi / 2) + pivot_point[1] * cos(pi / 2)))

            rotate(group, pi / 2)
            translate(group, vector=vector_translate)

        # Orientation.
        if self.orientation == "left":
            mirror(objects=group, mirror_type="y")

            if self.direction == "vertical":
                translate(group, vector=(self.box_height + x * 2, 0))
            else:
                if longitudinal:
                    translate(group, vector=(self.reinforcement_length + x * 2, 0))
                else:
                    translate(group, vector=(self.box_width + x * 2, 0))
